//
// Title:   SFLFileFormat.cpp
// Created: Tue Sep 24 18:28:23 2002
// Authors: Oliver Knoll, Mark Pauly, Tim Weyrich
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <memory.h>

#include <qstring.h>
#include <qfile.h>
#include <qapplication.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <assert.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Utilities/UserMessages/src/StatusBar.h"
#include "../ui/SRFCustomizeDialog.h"
#include "SRFFileFormat.h"

using namespace std;
	
static const unsigned int BASE_COUNTER = 50000; ///< an artefact, for having visualisation of the loading.

// **************
// public methods
// **************

SRFFileFormat::SRFFileFormat() : storeRadius (true) {}

SRFFileFormat::~SRFFileFormat(){}

bool SRFFileFormat::isReadable (const FileHeader fileHeader) {
  int spaceFind = (QString (fileHeader)).find (QChar (' '));
  if (fileHeader[0] == '#' || spaceFind != -1)
	return true;
  return false;
}

bool SRFFileFormat::readFromFile (const QString filePath, const bool mergeWithCurrent) {
  StatusBar * statusBar = StatusBar::getInstance (); // informations for users
  Scene * scene = Scene::getInstance ();
  Object * object;

  if (filePath.isNull () || filePath.isEmpty())
	return false;
  
  ifstream input (filePath, ios::in);
    
  if (!input) { 
	statusBar->showMessage (" Error : Opening file " + filePath, 0);
    return false; 
  } 
  
  statusBar->showMessage ("Loading, please wait... ", 0);
  
  if (!mergeWithCurrent) {
	scene->reset (true, false);
	scene->setAutoDelete (true);
  }
 
  
  object = new Object ("srfObject");
  SurfelCollection * surfelList = object->getSurfelCollection ();
  
  float surfelRadius = readHeader (input);
  unsigned int cpt = 0;
  
  while (!input.eof ()) {
	float x, y, z, nx, ny, nz, r, g, b;
	SurfelInterface * currentSurfel;
	QColor currentColor;
	input >> x >> y >> z >> nx >> ny >> nz >> r >> g >> b;
	currentSurfel = surfelList->addSurfel (false);
	currentColor.setRgb ((int)((unsigned char)(255.0 * r)), // we must ensure
						 (int)((unsigned char)(255.0 * g)), // that the value will
						 (int)((unsigned char)(255.0 * b)));// not exceed 255
	scene->setSurfelPosition (currentSurfel, Vector3D (x, y, z), false);
	scene->setSurfelNormal (currentSurfel, Vector3D (nx, ny, nz), false);
	scene->setSurfelRadius (currentSurfel, surfelRadius, false);
	scene->setSurfelDiffuseColor (currentSurfel, currentColor.rgb (), false);
	cpt ++;
	if (cpt % (BASE_COUNTER / 100) == 0)
      statusBar->setProgress (((int)((float)cpt / (float)BASE_COUNTER * 100.0f) % BASE_COUNTER), 
							  false);
  } 
  statusBar->setProgress (100, false);
  
  scene->addObject (object, true, true);
    
  input.close ();
  statusBar->showMessage (filePath + " Loaded!", 0);
  
  return true;	
}

bool SRFFileFormat::writeToFile (const QString filePath, const bool onlyActiveObject) {
  Scene * scene = Scene::getInstance ();
  StatusBar * statusBar = StatusBar::getInstance (); // informations for users
  Object * object;
  ofstream output (filePath);
 
  if (!output) { 
	statusBar->showMessage (" Error : Opening file " + filePath, 0);
    return false; 
  } 
  
  statusBar->showMessage ("Saving, please wait... ", 0);

  if (storeRadius) {
	object = scene->getFirstObject ();
	float baseRadius = defineAverageRadius (object);
  	output << "#" << baseRadius << endl;
  } 
  
  if (onlyActiveObject)
	writeObjectToStream (scene->getActiveObject (), output);
  else {
	for (object = scene->getFirstObject ();
		 object != NULL; 
		 object = scene->getNextObject ())
	  writeObjectToStream (object, output); 

  return true;
  }
}
  
QString SRFFileFormat::getFileFilter () const {
	return QString ("SRF Format (*.srf)");
}

QString SRFFileFormat::getFileSuffix () const {
	return QString ("srf");
}

void SRFFileFormat::customize (QWidget *parent) {

	SRFCustomizeDialog *srfCustomizeDialog;

	srfCustomizeDialog = new SRFCustomizeDialog (parent, "customizeDialog", true);
	// initialize
	srfCustomizeDialog->storeRadiusCheckBox->setChecked (true);
	
	srfCustomizeDialog->exec();
	
	storeRadius = srfCustomizeDialog->storeRadiusCheckBox->isChecked();

	delete srfCustomizeDialog;
}

// ***************
// private methods
// ***************

static const unsigned int MAX_COMMENT = 256;
static char dummy[MAX_COMMENT];

float SRFFileFormat::readHeader (ifstream& input) {
  char c;
  input.get (c);
  input.putback (c); // back to the stream
  if (c != '#') // no comment
    return DEFAULT_RADIUS; 
  
  input.get (c);
  float res; 
  input >> res;
  input.getline (dummy, MAX_COMMENT);
  
  // read the rest of the com.
  input.get (c);
  input.putback (c); // back to the stream
  while (c == '#') {
    input.getline (dummy, MAX_COMMENT);
    input.get (c);
    input.putback (c); // back to the stream
  }
  return res;
}

void SRFFileFormat::writeObjectToStream (Object * object, ofstream & output) {
  SurfelInterface * surfel;
  SurfelCollection * collection = object->getSurfelCollection ();
  unsigned int i, nofSurfels = collection->getNofSurfels ();
  StatusBar * statusBar = StatusBar::getInstance (); 
 
  statusBar->showMessage ("Saving object in srf...", 0);
  for (i = 0, surfel = collection->getFirstSurfel ();
       i < nofSurfels; 
       i++, surfel = collection->getNextSurfel ()) {
	Vector3D pos = surfel->getPosition ();
	Vector3D norm = surfel->getNormal ();
	QColor color = QColor (surfel->getDiffuseColor ());
	output << pos[0] << " " << pos[1] << " " << pos[2] << endl;
	output << norm[0] << " " << norm[1] << " " << norm[2] << endl;
	output << (float)color.red () / 255.0 << " " 
		   << (float)color.green () / 255.0 << " " 
		   << (float)color.blue () / 255.0 << endl << endl;
	if (i % (nofSurfels / 100) == 0)
      statusBar->setProgress ((int)((float)i / (float)nofSurfels * 100.0f), false);
  }
  statusBar->setProgress (100, false);
}

float SRFFileFormat::defineAverageRadius (Object * object) {
  assert (object != NULL);
  
  SurfelCollection * collection = object->getSurfelCollection ();
  unsigned int i, nofSurfels = collection->getNofSurfels ();
  float acc = 0.0;
  SurfelInterface * surfel;  
  
  for (i = 0, surfel = collection->getFirstSurfel ();
	   i < nofSurfels; 
       i++, surfel = collection->getNextSurfel ())
	acc += surfel->getRadius ();
  
  return acc/nofSurfels;
}

// ****************
// access functions
// ****************

FileFormatPluginInterface *createFileFormatPlugin() {
	return new SRFFileFormat();
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
